import aiohttp
import logging
from typing import List, Dict

from aiohttp.web_exceptions import HTTPForbidden

from feed_manager.engine.errors import NotPermissioned

_logger = logging.getLogger(__name__)


class DelegatesPermissionError(NotPermissioned):
    def __init__(self):
        super(DelegatesPermissionError, self).__init__(
            f"Auth token not permissioned for delegations request"
        )


class DelegatesClient:
    def __init__(self, config):
        self.client_session = aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(
                total=config.getint("delegatesclient", "timeout", fallback=60)
            )
        )

    async def close(self):
        await self.client_session.close()

    async def get_delegates(self, base_url: str, auth_token: str) -> List[str]:
        try:
            delegates_json = await self.make_request(base_url, auth_token)
        except HTTPForbidden:
            raise DelegatesPermissionError()

        return [
            delegate["orgKey"] for delegate in delegates_json.get("organizations", [])
        ]

    async def make_request(self, base_url: str, auth_token: str) -> Dict:
        async with self.client_session.get(
            f"{base_url}/appservices/v1/delegations",
            headers={"X-Auth-Token": auth_token},
        ) as resp:
            resp.raise_for_status()

            return await resp.json()
