import asyncio
import logging
import traceback
from pscr_common.stream import stream
from typing import AnyStr, Mapping

from yaml import dump as yamldump

from feed_manager.engine.feed_engine import FeedEngine
from feed_manager.engine.report_engine import ReportNotFound, ReportNotPermissioned
from feed_manager.engine.watchlist_engine import WatchlistEngine
from feed_manager.utils import aws_client
from feed_manager.utils.mem_cache import MemCache
from feed_manager.utils.config import Config
from feed_manager.utils.publisher import Publisher
import aioboto3
from feed_manager.utils.tasks.background_task import BackgroundTask

_logger = logging.getLogger(__name__)
MINUTES = 60
HOURS = 3600
EMPTY_ORG_FILE_SIZE = len('{"watchlists":[],"reports":[],"hash":"NA"}'.encode("utf-8"))


class DeprovisioningTask(BackgroundTask):
    def __init__(
        self,
        feed_engine: FeedEngine,
        watchlist_engine: WatchlistEngine,
        publisher: Publisher,
    ):
        super(DeprovisioningTask, self).__init__(
            name="deprovisioning_task",
            loop_sec=5 * MINUTES,
            sleep_first=True,
            execute_once=False,
        )
        self.cache = MemCache()
        self.reset_cache()
        self.feed_engine = feed_engine
        self.watchlist_engine = watchlist_engine
        self.publisher = publisher

    def reset_cache(self) -> None:
        self.cache.set("orgs_to_deprovision", list())

    async def execute(self, app) -> None:
        orgs = set(self.cache.get("orgs_to_deprovision"))
        if not orgs:
            _logger.debug("No orgs to deprovision.")
            return
        _logger.info(f"Deprovisioning the following orgs: {', '.join(orgs)}")
        statuses = await asyncio.gather(
            *(self._deprovision(org_key) for org_key in orgs)
        )
        _logger.info(
            f"Finished {len(orgs)} deprovisioning tasks.\n{yamldump(statuses)}"
        )
        self.reset_cache()

    async def _deprovision(self, org_key: AnyStr) -> Mapping[AnyStr, AnyStr]:
        status = "Unknown"
        try:
            reports = []
            # Collect Reports from feeds and watchlists
            feeds = await self.feed_engine.get_feeds(org_key, include_public=False)
            if not feeds:
                feeds = tuple()
            for feedinfo in feeds:
                f_reports = await self.feed_engine.get_feed_reports(
                    org_id=org_key, feed_id=feedinfo.id, remove_id_prefix=False
                )
                if f_reports:
                    reports += stream(f_reports).map(lambda x: x.id).to_list()
            watchlists = await self.watchlist_engine.get_watchlists(org_key)
            if not watchlists:
                watchlists = tuple()
            for watchlist in watchlists:
                if watchlist.report_ids:
                    reports += watchlist.report_ids
            # Delete Reports and IOC
            report_engine = self.watchlist_engine.report_engine
            deleted_reports = list()
            for r in reports:
                try:
                    await report_engine.verify_report_permissions(org_key, r)
                except (ReportNotFound, ReportNotPermissioned):
                    continue
                async with self.watchlist_engine.db_pool.rw_transaction() as conn:
                    await report_engine.delete_report(conn, r)
                    deleted_reports.append(r)
            # Notify feed-search
            await self.publisher.publish_report_notifications(deleted_reports)

            # Delete Feeds
            for feedinfo in feeds:
                await self.feed_engine.remove_feed(org_key, feedinfo.id)
            # Delete Watchlists
            await self.watchlist_engine.remove_all_watchlists(org_key)
            status = "Success"
            self.cache.set("deprovisioning_last_error", None)
        except Exception as e:
            _logger.error(
                f"Error deprovisioning {org_key}: {e}\n{traceback.format_exc()}"
            )
            self.cache.set("deprovisioning_last_error", repr(e))
            status = "Failure"
        return {"org_key": org_key, "status": status}


class S3DeprovisioningTask(BackgroundTask):
    def __init__(self, config: Config):
        super(S3DeprovisioningTask, self).__init__(
            name="s3_deprovisioning_task",
            loop_sec=12 * HOURS,
            sleep_first=False,
            execute_once=False,
        )
        self.config = config
        self.s3_org_files_bucket = config.get("watchlist", "augmentation_bucket")

    async def execute(self, app) -> None:
        # Delete ingress/query detector org files
        prefixes = set()
        async with aws_client(self.config, "s3") as s3:
            paginator = s3.get_paginator("list_objects")
            async for rsp in paginator.paginate(
                Bucket=self.s3_org_files_bucket, Prefix="watchlists/org=", Delimiter="/"
            ):
                for prefix in rsp.get("CommonPrefixes", []):
                    prefixes.add(prefix["Prefix"])

            for prefix in prefixes:
                rsp = await s3.list_objects_v2(
                    Bucket=self.s3_org_files_bucket, Prefix=prefix
                )
                empty = True
                contents = rsp.get("Contents", [])
                if not contents:
                    _logger.warning(
                        f"Malformed response {rsp} for prefix {prefix}. Skipping."
                    )
                    continue
                for obj in contents:
                    if obj["Key"].endswith(".json"):
                        if int(obj["Size"]) > EMPTY_ORG_FILE_SIZE:
                            empty = False
                            break
                    if obj["Key"].endswith(".json.dirty"):
                        if int(obj["Size"]) > 0:
                            empty = False
                            break
                if empty:
                    objects = (
                        stream(rsp.get("Contents", []))
                        .map(lambda x: x["Key"])
                        .to_list()
                    )
                    await s3.delete_objects(
                        Bucket=self.s3_org_files_bucket,
                        Delete={"Objects": [{"Key": o} for o in objects]},
                        ChecksumAlgorithm="SHA256",
                    )
                    _logger.info(f"Deleted the following empty org-files: {objects}")
