import logging

from aio_pyformance.registry import MetricsRegistry

from feed_manager.utils.db import ReadWriteDbPool

from feed_manager.engine.watchlist_engine import WatchlistEngine
from pscr_common.utils import get_utcnow_with_tzinfo

from feed_manager.utils.tasks.background_task import BackgroundTask
from feed_manager.utils.config import Config

_logger = logging.getLogger(__name__)


class OrgWatchlistDirtyTask(BackgroundTask):
    def __init__(
        self,
        config: Config,
        db_pool: ReadWriteDbPool,
        metrics: MetricsRegistry,
        watchlist_engine: WatchlistEngine,
    ):
        super(OrgWatchlistDirtyTask, self).__init__(
            "org_watchlist_dirty",
            loop_sec=config.getint("watchlist", "augment_check_sec", fallback=30),
            sleep_first=False,
            execute_once=False,
        )
        self.db_pool = db_pool
        self.metrics = metrics
        self.watchlist_engine = watchlist_engine

    async def execute(self, app):
        dirty_tuple = await self.watchlist_engine.get_dirty_org_for_build(
            get_utcnow_with_tzinfo()
        )

        while dirty_tuple:
            _logger.info("Dirty watchlist org: %s", dirty_tuple.org_id)

            await self.metrics.histogram("watchlists.dirty.queue.duration").add(
                dirty_tuple.queue_time_sec()
            )

            with self.metrics.timer("watchlists.dirty.build").time():
                watchlists = await self.watchlist_engine.get_org_augmented_watchlists_by_match_type(
                    dirty_tuple.org_id,
                    {
                        "query": lambda ioc: ioc.match_type == "query",
                        "ingress": lambda ioc: ioc.match_type != "query",
                    },
                )

            for match_type, augmented in watchlists.items():
                with self.metrics.timer(
                    f"watchlists.dirty.publish.{match_type}"
                ).time():
                    await self.watchlist_engine.publish_watchlists(
                        dirty_tuple.org_id,
                        dirty_tuple.rebuild_timestamp,
                        match_type,
                        augmented,
                    )

            await self.watchlist_engine.set_org_watchlist_buildtime(
                dirty_tuple.org_id, dirty_tuple.rebuild_timestamp
            )

            dirty_tuple = await self.watchlist_engine.get_dirty_org_for_build(
                get_utcnow_with_tzinfo()
            )
